#include "Preconditioner.h"
#include <algorithm>

namespace AMDiS {

  template<typename VectorType>
  GMResSolver<VectorType>::GMResSolver(::std::string name)
    : OEMSolver<VectorType>(name),
      TOL(1.e-25), 
      restart(0), 
      dim(0), 
      r(NULL), 
      v(NULL), 
      U(NULL), 
      LR(NULL), 
      givens(NULL), 
      w(NULL), 
      y(NULL), 
      gmres_k_residual_0(0.0)
  {
    FUNCNAME("GMResSolver::GMResSolver");
    GET_PARAMETER(0, name + "->restart", "%d", &restart);
    TEST_EXIT(restart)("restart not set\n");
  }

  template<typename VectorType>
  GMResSolver<VectorType>::~GMResSolver()
  {}

  template<typename VectorType>
  void GMResSolver<VectorType>::init()
  {
    r  = this->vectorCreator->create();
    
    dim = size(r);
    restart = ::std::max(0, ::std::min(restart, dim));

    v  = this->vectorCreator->create();
    U  = GET_MEMORY(VectorType*, restart);
    LR = GET_MEMORY(double*, restart);
    givens = GET_MEMORY(double*, restart);

    for (int i = 0; i < restart; i++) {
      U[i] = this->vectorCreator->create();
      LR[i] = GET_MEMORY(double, restart);
      givens[i] = GET_MEMORY(double, 2);
    }
    
    w = GET_MEMORY(double, restart);
    y = GET_MEMORY(double, restart);
  }

  template<typename VectorType>
  void GMResSolver<VectorType>::exit()
  {
    if (r) {
      int dim = restart; 
      int k   = ::std::max(0,::std::min(restart, dim));

      this->vectorCreator->free(r);
      this->vectorCreator->free(v);
    
      for (int i = 0; i < k; i++) {
	this->vectorCreator->free(U[i]);
	FREE_MEMORY(LR[i], double, k);
	FREE_MEMORY(givens[i], double, 2);
      }
      FREE_MEMORY(U, VectorType*, k);
      FREE_MEMORY(LR, double*, k);
      FREE_MEMORY(givens, double*, k);

      FREE_MEMORY(w, double, k);
      FREE_MEMORY(y, double, k);
    }
  }

  template<typename VectorType>
  int GMResSolver<VectorType>::solve_k(MatVecMultiplier<VectorType> *matVec,
				       VectorType *x, const VectorType *b)
  {
    FUNCNAME("GMResSolver::solve_k");

    VectorType *um1 = NULL;
    double  c, s, wm1;

    int k = ::std::max(0, ::std::min(restart, dim));

    /*--------------------------------------------------------------------------*/
    /*  Initialization                                                          */
    /*--------------------------------------------------------------------------*/

    // r = Ax
    matVec->matVec(NoTranspose, *x, *r);

    for (int i = 0; i < dim; i++) {
      (*r)[i] -= (*b)[i];
      (*r)[i] *= -1.0;
    }

    if (this->leftPrecon)
      this->leftPrecon->precon(r);

    gmres_k_residual_0 = norm(r);
    if (gmres_k_residual_0 < this->tolerance) {
      this->residual = gmres_k_residual_0;
      return(0);
    }

    /*--------------------------------------------------------------------------*/
    /*  construct k-dimensional Krylov space                                    */
    /*--------------------------------------------------------------------------*/

    wm1 = householderVector(gmres_k_residual_0, U[0], r, 0);
    um1 = U[0];

    double rj, sigma, maxi, val;
    VectorType *uj;
    int m = 0;

    for (m = 0; m < k; m++) {
      w[m] = wm1;

      newBasisVector(m + 1, k, U, LR, matVec, r, v, y);

      if (m + 1 < dim) {
	double norm = 0.0;
	for (int i = m + 1; i < dim; i++) {
	  norm += (*r)[i] * (*r)[i];  
	}
	norm = sqrt(norm);
	
	if (norm > TOL) {
	  /****************************************************************************/
	  /*  one of the last components of r is not zero; calculate Householder      */
	  /*  vector; if m < k-1, we need the Householder vector for the calculation  */
	  /*  of the next basis vector => store it; if m == k-1 we do not need this   */
	  /*  vector anymore => do not store it!                                      */
	  /****************************************************************************/
	  if (m < k - 1) {
	    um1 = U[m + 1];
	    (*r)[m + 1] = householderVector(norm, um1, r, m + 1);
	  } else {
	    (*r)[m + 1] = householderVector(norm, NULL, r, m + 1);
	  }
	}
      }

      for (int j = 0; j < m; j++) {
	rj = (*r)[j];   

	c = givens[j][0];
	s = givens[j][1];
	
	(*r)[j]   =  c * rj + s * (*r)[j + 1];
	(*r)[j+1] = -s * rj + c * (*r)[j + 1];
      }

      if (m + 1 < dim  &&  abs((*r)[m + 1]) > TOL) {
	/****************************************************************************/
	/*  Find Givens rotation J_m such that,                                     */
	/*       a)   (J_m r)[i] = r[i],  i < m,                                    */
	/*       b)   (J_m r)[m+1] = 0.0                                            */
	/*         => (J_m r)[m] = +- sqrt(r[m]^2 + r[m+1]^2) =: sigma              */
	/*                                                                          */
	/*                                                                          */
	/*                     |1  0  .  .  .  .  0|                                */
	/*                     |0  1  .  .  .  .  .|                                */
	/*                     |.     .           .|          c = r[m]/sigma        */
	/*            J_m =    |.        .        .|          s = r[m+1]/sigma      */
	/*                     |.           .     .|                                */
	/*                     |.              c  s| m                              */
	/*                     |0  .  .  .  . -s  c| m+1                            */
	/*                                     m  m+1                               */
	/****************************************************************************/

	maxi = ::std::max((*r)[m], (*r)[m+1]);

	c = (*r)[m] / maxi;
	s = (*r)[m + 1] / maxi;
	sigma = maxi * sqrt(c * c + s * s);
	if ((*r)[m] < 0)
	  sigma = -sigma;
	givens[m][0] = c = (*r)[m]/sigma;
	givens[m][1] = s = (*r)[m+1]/sigma;
	
	(*r)[m] = sigma;  /* r[m+1] == 0   automatically!                     */
	
	wm1 = -s * w[m];  /*  |wm1| is the new residual!                        */
	w[m] *= c;
      } else {
	wm1 = 0.0;
      }

      /****************************************************************************/
      /*  now, store the first m components of the column vector r in the the mth */
      /*  column of LR                                                            */
      /****************************************************************************/
      
      for (int j = 0; j <= m; j++)
	LR[j][m] = (*r)[j];

      /****************************************************************************/
      /*  test, whether tolarance is reached or not                               */
      /****************************************************************************/

      if (abs(wm1) < this->tolerance) {
	m++;
	break;
      }

      /****************************************************************************/
      /*  tolerance not reached: calculate and store (m+1)th row of matrix L;     */
      /*  this row is only needed for the computation of the (m+1)th column of    */
      /*  the orthogonal matrix Q_m; this vector is the additional basis vector   */
      /*  for the enlargement of the Krylov space; only needed for m < k-1        */
      /*  L[m+1][j] = <u_(m+1),u_j>_2                                             */
      /*  (m+1)th vector u = umi is allready stored at U+(m+1)                    */
      /****************************************************************************/
      
      if (m < k-1) {
	for (int j = 0; j < m + 1; j++) {
	  uj = U[j];
	  val = 0.0;
	  
	  for (int i = m + 1; i < dim; i++) {
	    val += (*um1)[i] * (*uj)[i];
	  }
	  
	  LR[(m+1)][j] = 2.0 * val;
	}
      }
    }

    /****************************************************************************/
    /*  and now solve the upper triangular system                               */
    /****************************************************************************/
    
    y[m - 1] = w[m - 1] / LR[m - 1][m - 1];  /* = LR[(m-1)*k+(m-1)]!                 */
    for (int j = m - 2; j >= 0; j--) {
      double yj = w[j];
      
      for (int l = j + 1; l < m; l++) {
	yj -= LR[j][l] * y[l];
      }
      
      y[j] = yj / LR[j][j];
    }

    /****************************************************************************/
    /*  calculate v = 2 U_m^T [e_0,....,e_m-1] y                                */
    /****************************************************************************/
  
    for (int i = 0; i < m; i++) {
      val = 0.0;
      
      for (int j = i; j < m; j++) {
	val += (*(U[i]))[j] * y[j];
      }

      (*v)[i] = 2.0 * val;
    }

    /****************************************************************************/
    /*  now solve L_m^T w = v in R^m  (upper triagonal system, diagonal == 1.0  */
    /****************************************************************************/

    w[m - 1] = (*v)[m - 1];
    for (int j = m - 2; j >= 0; j--) {
      double wj = (*v)[j];
      
      for (int l = j + 1; l < m; l++) {
	wj -= LR[l][j]*w[l];
      }

      w[j] = wj;
    }

    /****************************************************************************/
    /*  v = [e_0,....,e_m-1] y - U_m w                                          */
    /****************************************************************************/

    for (int i = 0; i < dim; i++) {
      double vi = 0.0;
      for (int j = 0; j < ::std::min(i + 1,m); j++)
	vi += (*(U[j]))[i] * w[j];
      (*v)[i] = -vi;
    }

    for (int j = 0; j < m; j++)
      (*v)[j] += y[j];

    /****************************************************************************/
    /*  and now, make the update of u :-)                                       */
    /****************************************************************************/

    if (this->rightPrecon)
      this->rightPrecon->precon(v);

    for (int i = 0; i < dim; i++)
      (*x)[i] += (*v)[i];

    this->residual = abs(wm1);
    return(m);
  }

  template<typename VectorType>
  int GMResSolver<VectorType>::solveSystem(MatVecMultiplier<VectorType> *matVec,
					   VectorType *x, VectorType *b)
  {
    FUNCNAME("GMResSolver::solveSystem()");
  
    double    old_res = -1.0;

    if (norm(b) < TOL) {
      INFO(this->info, 2)("b == 0, x = 0 is the solution of the linear system\n");
      setValue(*x, 0.0);
      this->residual = 0.0;
      return(0);
    }
 
    START_INFO();
    for (int iter = 0; iter <= this->max_iter; iter++) {
      int k = solve_k(matVec, x, b);
      if (SOLVE_INFO(iter, this->residual, &old_res))
	return(iter);
      TEST_EXIT(k != 0)("this must not happen\n");
    }
  
    ERROR_EXIT("Should not be reached\n");
    return(0);
  }

  template<typename VectorType>
  double GMResSolver<VectorType>::householderVector(double sigma,
						    VectorType *u,
						    VectorType *x,
						    int offset)
  {
    FUNCNAME("GMResSolver::householderVector()");

    if (dim <= 0) {
      MSG("dim <= 0\n");
      return(0.0);
    }

    if ((*x)[offset] >= 0) {
      if (u) {
	double beta = sqrt(0.5 / (sigma * (sigma + (*x)[offset])));
	(*u)[offset] = beta * ((*x)[offset] + sigma);

	for (int i = offset + 1; i < dim; i++)
	  (*u)[i] = beta * (*x)[i];
      }
      return(-sigma);

    } else {
      if (u) {
	double beta = sqrt(0.5 / (sigma * (sigma - (*x)[offset])));
	(*u)[offset] = ((*x)[offset] - sigma) * beta;
    
	for (int i = offset + 1; i < dim; i++)
	  (*u)[i] = (*x)[i] * beta;
      }

      return(sigma);
    }
  }

  template<typename VectorType>
  void GMResSolver<VectorType>::newBasisVector(int m, int k, VectorType **U, 
					       double **LR,  
					       MatVecMultiplier<VectorType> *matVec,
					       VectorType *v, 
					       VectorType *b, double* y)
  {
    /****************************************************************************/
    /*  first calculate y = 2 U_m^T e_m                                         */
    /****************************************************************************/
  
    for (int j = 0; j < m; j++)
      (*b)[j] = 2.0*(*(U[j]))[m-1];


    /****************************************************************************/
    /*  now solve L_m^T y = b in R^m  (upper triagonal system, diagonal == 1.0  */
    /****************************************************************************/

    y[m-1] = (*b)[m-1];
    for (int j = m-2; j >= 0; j--) {
      double yj = (*b)[j];

      for (int l = j + 1; l < m; l++)
	yj -= LR[l][j]*y[l];

      y[j] = yj;
    }
  
    /****************************************************************************/
    /*  b = -U_m y + e_m                                                        */
    /****************************************************************************/

    for (int i = 0; i < dim; i++) {
      double bi = 0.0;
      
      for (int j = 0; j < ::std::min(i + 1,m); j++)
	bi += (*(U[j]))[i] * y[j];
      
      (*b)[i] = -bi;
    }
    (*b)[m - 1] += 1.0;

    /****************************************************************************/
    /*  v = Ab                                                                  */
    /****************************************************************************/

    if (this->rightPrecon)
      this->rightPrecon->precon(b);

    matVec->matVec(NoTranspose, *b, *v);
    
    if (this->leftPrecon)
      this->leftPrecon->precon(v);

    /****************************************************************************/
    /*  b = 2 U_m^T v in R^m                                                    */
    /****************************************************************************/

    for (int j = 0; j < m; j++) {
      double bj = 0.0;

      for (int i = j; i < dim; i++)
	bj += (*(U[j]))[i]*(*v)[i];
      (*b)[j] = 2.0*bj;
    }

    /****************************************************************************/
    /*  now solve L_m y = b in R^m  (lower triagonal system, diagonal == 1.0    */
    /****************************************************************************/

    y[0] = (*b)[0];
    for (int j = 1; j < m; j++) {
      double yj = (*b)[j];
      
      for (int l = 0; l < j; l++)
	yj -= LR[j][l] * y[l];
      
      y[j] = yj;
    }

    /****************************************************************************/
    /*  v = v - U_m y                                                           */
    /****************************************************************************/

    for (int i = 0; i < dim; i++) {
      double vi = 0.0;
      
      for (int j = 0; j < ::std::min(i + 1,m); j++)
	vi += (*(U[j]))[i]*y[j];
      
      (*v)[i] -= vi;
    }

    return;
  }
}
