#include <fstream>

#include "MacroWriter.h"
#include "DataCollector.h"
#include "Mesh.h"
#include "DOFAdmin.h"
#include "FiniteElemSpace.h"
#include "DOFVector.h"
#include "DOFIterator.h"
#include "ElInfo.h"
#include "SurfaceRegion_ED.h"
#include "ElementRegion_ED.h"
#include <string>
#include "Traverse.h"

namespace AMDiS {

  DOFVector< ::std::list<VertexInfo> >* MacroWriter::vertexInfos = NULL;
  ::std::list<ElementInfo> MacroWriter::elements;
  
  Mesh *MacroWriter::mesh                          = NULL;
  FILE *MacroWriter::macroFile                     = NULL;
  FILE *MacroWriter::periodicFile                  = NULL;
  int   MacroWriter::n0                            = 0;
  int   MacroWriter::nv                            = 0;
  int   MacroWriter::ne                            = 0;
  int   MacroWriter::nc                            = 0;
  int   MacroWriter::dim                           = 0;

  ::std::map<int, int> MacroWriter::outputIndices;
  
  ::std::vector<DimVec<bool> > MacroWriter::periodicConnections;

  bool (*MacroWriter::writeElement)(ElInfo*) = NULL;
  
  int MacroWriter::writeMacro(DataCollector *dc,
			      const char *name, 			     
			      double time,
			      int level,
			      Flag traverseFlag,
			      bool (*writeElem)(ElInfo*))
  {
    FUNCNAME("MacroWroter::writeFile");
    TEST_EXIT(dc)("no data collector\n");

    writeElement = writeElem;

    ::std::ofstream file;
    ::std::list<ElementInfo> *elements = dc->getElementInfos();
    DOFVector< ::std::list<VertexInfo> > *vertexInfos = dc->getVertexInfos();

    int dow = Global::getGeo(WORLD);
    int dim = dc->getMesh()->getDim();
    int nv = dc->getNumberVertices();
    int ne = dc->getNumberElements();
    int vertices = dc->getMesh()->getGeo(VERTEX);

    file.open(name);

    // === print file header ===
    file << "mesh name: " << dc->getMesh()->getName() << ::std::endl << ::std::endl;
    file << "time: " << ::std::scientific << time << ::std::endl << ::std::endl;
    
    file << "DIM: " << dim << ::std::endl;
    file << "DIM_OF_WORLD: " << dow << ::std::endl << ::std::endl;

    file << "number of vertices: " << nv << ::std::endl;
    file << "number of elements: " << ne << ::std::endl << ::std::endl;

    // === print vertex coords and remember global output indices ===
    file << "vertex coordinates:" << ::std::endl;

    DOFVector< ::std::list<VertexInfo> >::Iterator it(vertexInfos, USED_DOFS);
    int i, counter = 0;

    // for all DOFs
    for(it.reset(); !it.end(); ++it) {
      // for all vertex infos of this DOF
      ::std::list<VertexInfo>::iterator it2;
      for(it2 = it->begin(); it2 != it->end(); ++it2) {
	it2->outputIndex = counter++;
	for(i = 0; i < dow; i++) {
	  file << ::std::scientific << it2->coords[i] << " ";
	}
	file << ::std::endl;
      }
    }

    // === print element vertices ===
    file << ::std::endl << "element vertices:" << ::std::endl;

    // iterate the element list
    ::std::list<ElementInfo>::iterator elementIt;

    for(elementIt = elements->begin(); elementIt != elements->end(); ++elementIt) {
      // for all vertices
      for(i = 0; i < vertices; i++) {
	file << elementIt->vertexInfo[i]->outputIndex << " ";
      }
      file << ::std::endl;
    }

    // === print boundaries ===
    file << ::std::endl << "element boundaries:" << ::std::endl;

    for(elementIt = elements->begin(); elementIt != elements->end(); ++elementIt) {
      // for all vertices
      for(i = 0; i < vertices; i++) {
	file << elementIt->boundary[i] << " ";
      }
      file << ::std::endl;
    }
    
    // === print neighbours ===
    file << ::std::endl << "element neighbours:" << ::std::endl;

    for(elementIt = elements->begin(); elementIt != elements->end(); ++elementIt) {
      // for all vertices
      for(i = 0; i < vertices; i++) {
	file << elementIt->neighbour[i] << " ";
      }
      file << ::std::endl;      
    }

    // === print boundary projections ===
    file << ::std::endl << "projections:" << ::std::endl;

    for(elementIt = elements->begin(); elementIt != elements->end(); ++elementIt) {
      // for all vertices
      for(i = 0; i < vertices; i++) {
	if(elementIt->projection[i])
	  file << elementIt->projection[i]->getID() << " ";
	else
	  file << "0 ";
      }
      file << ::std::endl;
    }

    // === print element regions ===
    file << ::std::endl << "element region:" << ::std::endl;
    
    for(elementIt = elements->begin(); elementIt != elements->end(); ++elementIt) {
      file << elementIt->elementRegion << ::std::endl;
    }

    // === print surface regions ===
    file << ::std::endl << "surface region:" << ::std::endl;

    for(elementIt = elements->begin(); elementIt != elements->end(); ++elementIt) {
      for(i = 0; i < vertices; i++) {
	file << elementIt->surfaceRegions[i] << " ";
      }
      file << ::std::endl;
    }

    // === print element types if necessary ===
    if (dim == 3) {
      file << ::std::endl << "element type:" << ::std::endl;
	
      for(elementIt = elements->begin(); elementIt != elements->end(); ++elementIt) {
	file << (int)(elementIt->type) << " ";
      }
    } 
  
    file.close();

    return(0);
  }


  void MacroWriter::writePeriodicFile(DataCollector *dc,
				      const ::std::string filename)
  {
    FUNCNAME("MacroWriter::writePeriodicFile2");
    TEST_EXIT(dc)("no data collector\n");
  
    ::std::ofstream file;

    file.open(filename.c_str());
  
    file << "associations: " << dc->getNumberConnections() << ::std::endl;
    file << ::std::endl << "mode  bc  el1 - local vertices <->  el2 - local vertices" << ::std::endl;

    ::std::list<PeriodicInfo>::iterator periodicIt;

    // Iterate on all periodic connections
    for(periodicIt = dc->getPeriodicInfos()->begin(); 
	periodicIt != dc->getPeriodicInfos()->end(); 
	++periodicIt) {

      ::std::map<int, int>::iterator mapIt;

      // Write mode and type of periodic connection
      file << periodicIt->mode << " " << periodicIt->type << " ";    

      // Write index of the first element
      file << periodicIt->outputIndex << " ";

      // Write local indices of the first element
      for (mapIt = periodicIt->vertexMap.begin();
	   mapIt != periodicIt->vertexMap.end();
	   ++mapIt) {
	file << mapIt->first << " ";
      }

      // Write index of the second element
      file << periodicIt->neighIndex << " ";

      // Write local indices of the second element
      for (mapIt = periodicIt->vertexMap.begin();
	   mapIt != periodicIt->vertexMap.end();
	   ++mapIt) {
	file << mapIt->second << " ";
      }

      file << ::std::endl;
    }
  
  
    file.close();
  }


  int MacroWriter::writeMacroOld(const FiniteElemSpace *fe_space, 
				 const char* name, double time,
				 int level,
				 Flag traverseFlag)
  {
    FUNCNAME("MacroWriter::writeMacro");
    TEST_EXIT(fe_space)("no feSpace\n");

    mesh = fe_space->getMesh();
    dim = mesh->getDim();
    int dow = Global::getGeo(WORLD);

    bool writeElType = (dim == 3);
    int vertices = mesh->getGeo(VERTEX);

    // === create filename ===
    ::std::string filename;
    if(name) {
      filename = ::std::string(name);
    } else {
      filename = mesh->getName();
    }
  
    // === open file ===
    if (!(macroFile = fopen(filename.c_str(), "w"))) {
      ERROR_EXIT("could not open file %s for writing\n", filename.c_str());
      return(0);
    }

    // === get admin and fe_space ===
    DOFAdmin* localAdmin = const_cast<DOFAdmin*>(fe_space->getAdmin());

    //   DOFAdmin* localAdmin = const_cast<DOFAdmin*>(mesh->getVertexAdmin());
    //   TEST_EXIT(localAdmin)("no dof admin for vertices\n");
    //   FiniteElemSpace *fe_space = 
    //     FiniteElemSpace::provideFESpace(localAdmin, NULL, NULL);

    // === create vertex info vector ===
    vertexInfos = NEW DOFVector< ::std::list<VertexInfo> >(fe_space, "vertex infos");

    // === delete old element list ===
    elements.resize(0, dim);

    n0 = localAdmin->getNumberOfPreDOFs(VERTEX);
    nv = ne = 0;

    // create output indices
    //mesh->traverse(level, traverseFlag, indexFct);

    TraverseStack stack;
    ElInfo *elInfo = stack.traverseFirst(mesh, level, traverseFlag);
    while(elInfo) {
      if(!writeElement || writeElement(elInfo)) indexFct(elInfo);
      elInfo = stack.traverseNext(elInfo);
    }

    traverseFlag |= 
      Mesh::FILL_NEIGH      | 
      Mesh::FILL_COORDS     |
      Mesh::FILL_OPP_COORDS |
      Mesh::FILL_BOUND;

    // === traverse mesh and fill vertex infos and element list ===
//     mesh->traverse(level, 
// 		   traverseFlag,
// 		   traverseFct);

    elInfo = stack.traverseFirst(mesh, level, traverseFlag);
    while(elInfo) {
      if(!writeElement || writeElement(elInfo)) traverseFct(elInfo);
      elInfo = stack.traverseNext(elInfo);
    }

    // === print file header ===
    fprintf(macroFile, "mesh name: %s\n\n", mesh->getName().c_str());
    fprintf(macroFile, "time: %e\n\n", time);

    fprintf(macroFile, "DIM: %d\n", dim);
    fprintf(macroFile, "DIM_OF_WORLD: %d\n\n", dow);

    fprintf(macroFile, "number of vertices: %d\n", nv);
    fprintf(macroFile, "number of elements: %d\n\n", ne);

    // === print vertex coords and remember global output indices ===
    fprintf(macroFile, "vertex coordinates:\n");

    DOFVector< ::std::list<VertexInfo> >::Iterator it(vertexInfos, USED_DOFS);
    int i, counter = 0;

    // for all DOFs
    for(it.reset(); !it.end(); ++it) {
      // for all vertex infos of this DOF
      ::std::list<VertexInfo>::iterator it2;
      for(it2 = it->begin(); it2 != it->end(); ++it2) {
	it2->outputIndex = counter++;
	for(i=0; i < dow; i++) {
	  fprintf(macroFile, "%e ", it2->coords[i]);
	}
	fprintf(macroFile, "\n");
      }
    }

    // === print element vertices ===
    fprintf(macroFile, "\nelement vertices:\n");

    // iterate the element list
    ::std::list<ElementInfo>::iterator elementIt;

    for(elementIt = elements.begin(); elementIt != elements.end(); ++elementIt) {
      // for all vertices
      for(i=0; i < vertices; i++) {
	fprintf(macroFile, "%d ", elementIt->vertexInfo[i]->outputIndex);
      }
      fprintf(macroFile, "\n");
    }

    // === print boundaries ===
    fprintf(macroFile, "\nelement boundaries:\n");

    for(elementIt = elements.begin(); elementIt != elements.end(); ++elementIt) {
      // for all vertices
      for(i=0; i < vertices; i++) {
	fprintf(macroFile, "%d ", elementIt->boundary[i]);
      }
      fprintf(macroFile, "\n");
    }

    // === print neighbours ===
    fprintf(macroFile, "\nelement neighbours:\n");

    for(elementIt = elements.begin(); elementIt != elements.end(); ++elementIt) {
      // for all vertices
      for(i=0; i < vertices; i++) {
	fprintf(macroFile, "%d ", elementIt->neighbour[i]);
      }
      fprintf(macroFile, "\n");
    }

    // === print boundary projections ===
    fprintf(macroFile, "\nprojections:\n");

    for(elementIt = elements.begin(); elementIt != elements.end(); ++elementIt) {
      // for all vertices
      for(i=0; i < vertices; i++) {
	if(elementIt->projection[i])
	  fprintf(macroFile, "%d ", elementIt->projection[i]->getID());
	else
	  fprintf(macroFile, "0 ");
      }
      fprintf(macroFile, "\n");
    }

    // === print element regions ===
    fprintf(macroFile, "\nelement region:\n");

    for(elementIt = elements.begin(); elementIt != elements.end(); ++elementIt) {
      fprintf(macroFile, "%d\n", elementIt->elementRegion);
    }

    // === print surface regions ===
    fprintf(macroFile, "\nsurface region:\n");

    for(elementIt = elements.begin(); elementIt != elements.end(); ++elementIt) {
      for(i=0; i < vertices; i++) {
	fprintf(macroFile, "%d ", elementIt->surfaceRegions[i]);
      }
      fprintf(macroFile, "\n");
    }

    // === print element types if necessary ===
    if (dim == 3) {
      if (writeElType) {
	fprintf(macroFile, "\nelement type:\n");

	for(elementIt = elements.begin(); elementIt != elements.end(); ++elementIt) {
	  fprintf(macroFile, "%d ", elementIt->type);
	}
      }
    }

    // === close file ===
    fclose(macroFile);
    macroFile = NULL;

    DELETE vertexInfos;

    return(1);
  }

  int MacroWriter::traverseFct(ElInfo* elInfo)
  {
    int i;
    const DegreeOfFreedom **dof = elInfo->getElement()->getDOF();
    DegreeOfFreedom vertexDOF;
    WorldVector<double> vertexCoords;

    // increment number of elements
    //ne++;

    // create ElementInfo
    ElementInfo elementInfo(dim);

    // read element region
    ElementData *ed = 
      elInfo->getElement()->getElementData(ELEMENT_REGION);
  
    if(ed) {
      elementInfo.elementRegion = 
	dynamic_cast<ElementRegion_ED*>(ed)->getRegion();
    }
    else {
      elementInfo.elementRegion = -1;
    }

    // read surface regions to element info
    ed = elInfo->getElement()->getElementData(SURFACE_REGION);

    elementInfo.surfaceRegions.set(-1);

    while(ed) {
      SurfaceRegion_ED *sr = dynamic_cast<SurfaceRegion_ED*>(ed);
      elementInfo.surfaceRegions[sr->getSide()] = sr->getRegion();
      ed = ed->getDecorated(SURFACE_REGION);
    }

    // for all vertices
    for(i=0; i < dim+1; i++) {
      // get coords of this vertex
      vertexCoords = elInfo->getCoord(i);

      // get dof index of this vertex
      vertexDOF = dof[i][n0];

      // search for coords at this dof
      ::std::list<VertexInfo>::iterator it =
	  find((*vertexInfos)[vertexDOF].begin(),
	       (*vertexInfos)[vertexDOF].end(),
	       vertexCoords);

      // coords not yet in list?
      if(it == (*vertexInfos)[vertexDOF].end()) {
	// create new vertex info
	VertexInfo newVertexInfo = {vertexCoords, -1};
      
	// add new vertex info to list
	(*vertexInfos)[vertexDOF].push_front(newVertexInfo);
      
	// set iterator to new vertex info
	it = (*vertexInfos)[vertexDOF].begin();

	// increment number of vertices
	nv++;
      }

      // fill element info
      elementInfo.vertexInfo[i] = it;
      elementInfo.boundary[i] = elInfo->getBoundary(i);
      elementInfo.projection[i] = elInfo->getProjection(i);
      elementInfo.neighbour[i] = 
	elInfo->getNeighbour(i) ?
	outputIndices[elInfo->getNeighbour(i)->getIndex()] :
	-1;
    }

    if(dim == 3) {
      elementInfo.type = (dynamic_cast<ElInfo3d*>(elInfo)->getType());
    }

    // remember element info
    elements.push_back(elementInfo);

    return 0;
  }

  void MacroWriter::writePeriodicFileOld(Mesh *aMesh, 
					 const ::std::string filename,
					 int level,
					 Flag traverseFlag)
  {
    // === open file ===
    if (!(periodicFile = fopen(filename.c_str(), "w"))) {
      ERROR_EXIT("could not open file %s for writing\n", filename.c_str());
    }

    mesh = aMesh;
    dim = mesh->getDim();

    // === create output indices and count connections ===

    n0 = mesh->getVertexAdmin()->getNumberOfPreDOFs(VERTEX);
    ne = 0;
    nc = 0;

    //mesh->traverse(level, traverseFlag, periodicFct1);

    TraverseStack stack;
    ElInfo *elInfo = stack.traverseFirst(mesh, level, traverseFlag);
    while(elInfo) {
      if(!writeElement || writeElement(elInfo)) periodicFct1(elInfo);
      elInfo = stack.traverseNext(elInfo);
    }

    nc /= 2;

    fprintf(periodicFile, "associations: %d\n", nc);
    fprintf(periodicFile, "\nmode  bc  el1 - local vertices <->  el2 - local vertices\n");
  
    traverseFlag |=
      Mesh::FILL_COORDS    |
      Mesh::FILL_OPP_COORDS|
      Mesh::FILL_NEIGH     | 
      Mesh::FILL_BOUND;
    

    //mesh->traverse(level, 
    //	   traverseFlag,
    //	   periodicFct2);

    elInfo = stack.traverseFirst(mesh, level, traverseFlag);
    while(elInfo) {
      if(!writeElement || writeElement(elInfo)) periodicFct2(elInfo);
      elInfo = stack.traverseNext(elInfo);
    }

    outputIndices.clear();
    periodicConnections.clear();

    // === close periodic file ===

    fclose(periodicFile);
    periodicFile = NULL;
  }

  int MacroWriter::periodicFct1(ElInfo *elInfo)
  {
    const Element *element = elInfo->getElement();

    outputIndices[element->getIndex()] = ne++;

    LeafDataPeriodic *ldp = dynamic_cast<LeafDataPeriodic*>
      (element->
       getElementData()->
       getElementData(PERIODIC));

    if(ldp) {
      nc += dynamic_cast<LeafDataPeriodic*>(ldp)->getInfoList().size();
    }

    periodicConnections.push_back(DimVec<bool>(dim, DEFAULT_VALUE, false));

    return 0;
  }

  int MacroWriter::periodicFct2(ElInfo *elInfo)
  {
    FUNCNAME("MacroWriter::periodicFct2");

    int i, j;

    const Element *element = elInfo->getElement(); 

    LeafDataPeriodic *ldp = dynamic_cast<LeafDataPeriodic*>
      (element->
       getElementData()->
       getElementData(PERIODIC));

    if(ldp) {
      ::std::list<LeafDataPeriodic::PeriodicInfo>::iterator it;
      for(it = dynamic_cast<LeafDataPeriodic*>(ldp)->getInfoList().begin();
	  it != dynamic_cast<LeafDataPeriodic*>(ldp)->getInfoList().end();
	  ++it) 
	{
	  int outputIndex = outputIndices[elInfo->getElement()->getIndex()];
	  int neighIndex  = outputIndices[elInfo->
					  getNeighbour(it->elementSide)->
					  getIndex()];

	  if(!periodicConnections[outputIndex][it->elementSide]) {

	    fprintf(periodicFile, "%d ", it->periodicMode);
	    fprintf(periodicFile, "%d ", it->type);


	    periodicConnections[outputIndex][it->elementSide] = true;
	    periodicConnections
	      [neighIndex][elInfo->getOppVertex(it->elementSide)] = true;

	    int index1, index2, dof1, dof2;
	    ::std::map<int, int> vertexMap;

	    for(i=0; i < dim; i++) {
	      index1 = 
		elInfo->getElement()->getVertexOfPosition(INDEX_OF_DIM(dim-1, dim),
							  it->elementSide,
							  i);
	      dof1 = elInfo->getElement()->getDOF(index1, n0);

	      for(j=0; j < dim; j++) {
		index2 = 
		  elInfo->getElement()->getVertexOfPosition(INDEX_OF_DIM(dim-1, dim),
							    elInfo->getOppVertex(it->elementSide),
							    j);
		dof2 = elInfo->getNeighbour(it->elementSide)->getDOF(index2, n0);

		if((dof1 == dof2) || (mesh->associated(dof1, dof2))) {
		  vertexMap[index1] = index2;
		  //MSG("%d ", outputIndices[elInfo->getElement()->getIndex()]); 

		  //	      ::std::cout << index1 << " " << index2 << " ";
		  break;
		}
	      }
	    }

	    fprintf(periodicFile, "%d ", outputIndex);
	    for(i=0; i < dim; i++) {
	      fprintf(periodicFile, "%d ", 
		      elInfo->
		      getElement()->
		      getVertexOfPosition(INDEX_OF_DIM(dim-1, dim),
					  it->elementSide,
					  i)
		      );
	    }
	    fprintf(periodicFile, "%d ", neighIndex);
	    for(i=0; i < dim; i++) {
	      fprintf(periodicFile, "%d", 
		      vertexMap[elInfo->
				getElement()->
				getVertexOfPosition(INDEX_OF_DIM(dim-1, dim),
						    it->elementSide,
						    i)]
		      );
	      fprintf(periodicFile, i < dim-1 ? " " : "\n");
	    }
	  }
	}
    }

    return 0;
  }

  int MacroWriter::indexFct(ElInfo *elInfo)
  {
    outputIndices[elInfo->getElement()->getIndex()] = ne++;  
    return 0;
  }

}
