#include <iostream>
#include <fstream>
#include <vector>
#include "time.h"
#include "MatVecMultiplier.h"
#include "FiniteElemSpace.h"
#include "DOFMatrix.h"
#include "DOFVector.h"
#include "MatrixVector.h"
#include "SystemVector.h"
#include "V3Vector.h"

namespace AMDiS {

  template<> void 
  StandardMatVec<DOFMatrix, DOFVector<double> >::matVec(MatrixTranspose transpose,
							const DOFVector<double> &x,
							DOFVector<double> &result,
							bool add)
  {  
    FUNCNAME("StandardMatVec::matVec()");

    TEST_EXIT((x.getFESpace() == matrix->getColFESpace()))
      ("finite element spaces of matrix and x must be the same!\n");

    mv(transpose, *matrix, x, result, add);
  }

  template <> void 
  StandardMatVec<Matrix<DOFMatrix*>, SystemVector>::matVec(MatrixTranspose transpose,
							   const SystemVector &x, 
							   SystemVector &result,
							   bool add)
  {
    mv(*matrix, x, result);
  }



  template<> void 
  StandardMatVec< ::std::vector< ::std::vector<MatEntry> >,
		  ::std::vector<double> >::matVec(MatrixTranspose transpose,
						  const ::std::vector<double> &x,
						  ::std::vector<double> &result,
						  bool add)
  {  
    FUNCNAME("StandardMatVec::matVec()");

    unsigned int nRows = matrix->size();

    TEST_EXIT(x.size() == nRows)("SIZE!\n");
    TEST_EXIT(result.size() == nRows)("SIZE!\n");

    if (!add) {
      for (unsigned int i = 0; i < nRows; i++) {
	result[i] = 0.0;
      }
    }

    for (unsigned int row = 0; row < nRows; row++) {
      for (unsigned int col = 0; col < (*matrix)[row].size(); col++) {
	MatEntry *m = &(*matrix)[row][col];
	
	result[row] += m->entry * x[m->col];
      }
    }
  }

  template<>
  void MassLumpingMatVec<DOFMatrix, DOFVector<double> >::calcMassLumpingMatrix(DOFMatrix *m)
  {
    // create new or clear old matrix
    if (!matrix) {
      matrix = NEW DOFMatrix(m->getRowFESpace(),
			     m->getColFESpace(),
			     "mass lumping matrix");
    }
    else {
      matrix->clear();
    }
  
    // for all rows
    DOFMatrix::Iterator rowIt1(m, USED_DOFS);
    DOFMatrix::Iterator rowIt2(matrix, USED_DOFS);
    for (rowIt1.reset(), rowIt2.reset(); 
	 !rowIt1.end(); 
	 ++rowIt1, ++rowIt2) 
      {
	double rowSum = 0.0;
	int rowSize = static_cast<int>(rowIt1->size());

	for (int i = 0; i < rowSize; i++) {
	  if ((*rowIt1)[i].col == DOFMatrix::NO_MORE_ENTRIES) {
	    break;
	  }
	  if ((*rowIt1)[i].col == DOFMatrix::UNUSED_ENTRY) {
	    continue;
	  }
      
	  double entry = (*rowIt1)[i].entry;

	  TEST_EXIT(entry >= 0)("entries must be >= 0\n");

	  rowSum += entry;
	}

	TEST_EXIT(rowSum > 1E-20)("matrix not inversable\n");

	(*rowIt2).resize(1);
	(*rowIt2)[0].col = rowIt2.getDOFIndex();
	(*rowIt2)[0].entry = 1.0 / rowSum;
      }
  }

  template<>
  void MassLumpingMatVec<Matrix<DOFMatrix*>, SystemVector>::calcMassLumpingMatrix(Matrix<DOFMatrix*> *m)
  {
    int numComponents = m->getNumRows();
    // create new or clear old matrix
    if(!matrix) {
      matrix = NEW Matrix<DOFMatrix*>(numComponents, numComponents);
      for (int i = 0; i < numComponents; i++) {
	(*matrix)[i][i] = NEW DOFMatrix((*m)[i][i]->getRowFESpace(),
					(*m)[i][i]->getColFESpace(),
					"mass lumping matrix");
      }
    }
    else {
      for (int i = 0; i < numComponents; i++)
	(*matrix)[i][i]->clear();
    }
  
    // for all rows
    for (int i = 0; i < numComponents; i++) {
      DOFMatrix::Iterator rowIt1((*m)[i][i], USED_DOFS);
      DOFMatrix::Iterator rowIt2((*matrix)[i][i], USED_DOFS);
      for (rowIt1.reset(), rowIt2.reset(); 
	   !rowIt1.end(); 
	   ++rowIt1, ++rowIt2) 
	{
	  double rowSum = 0.0;
	  int rowSize = static_cast<int>(rowIt1->size());

	  for (int j = 0; j < rowSize; j++) {
	    if((*rowIt1)[j].col == DOFMatrix::NO_MORE_ENTRIES) {
	      break;
	    }
	    if((*rowIt1)[j].col == DOFMatrix::UNUSED_ENTRY) {
	      continue;
	    }
      
	    double entry = (*rowIt1)[j].entry;

	    TEST_EXIT(entry >= 0)("entries must be >= 0\n");

	    rowSum += entry;
	  }

	  TEST_EXIT(rowSum > 1E-20)("matrix not inversable\n");

	  (*rowIt2).resize(1);
	  (*rowIt2)[0].col = rowIt2.getDOFIndex();
	  (*rowIt2)[0].entry = 1.0 / rowSum;
	}
    }
  }


  StandardMatVec<Matrix<DOFMatrix*>, SystemVector> _enf_(NULL);
  StandardMatVec<DOFMatrix, DOFVector<double> > _grmpf_(NULL);

}
