#include "QPInfo.h"
#include "Quadrature.h"
#include "DOFVector.h"

namespace AMDiS {

  ::std::map<const Quadrature*, QPInfo*> QPInfo::qpInfos_;

  QPInfo::QPInfo(const Quadrature *quad)
    : quadrature_(quad),
      currentElInfo_(NULL),
      coordsAtQPs_(NULL),
      coordsNumPointsValid_(0),
      elementNormalAtQPs_(NULL),
      elementNormalConst_(NULL),
      elementNormalNumPointsValid_(0),
      grdLambdaAtQPs_(NULL),
      grdLambdaConst_(NULL),
      grdLambdaNumPointsValid_(0)
  {
    numPoints_ = quadrature_->getNumPoints();
  }

  QPInfo::~QPInfo()
  {
    int i;

    if(coordsAtQPs_) DELETE [] coordsAtQPs_;

    if(elementNormalAtQPs_) {
      for(i = 0; i < numPoints_; i++) {
	DELETE elementNormalAtQPs_[i];
      }
      FREE_MEMORY(elementNormalAtQPs_,
		  WorldVector<double>*,
		  numPoints_);
    }

    if(elementNormalConst_) {
      FREE_MEMORY(elementNormalConst_,
		  WorldVector<double>*,
		  numPoints_);
    }

    if(grdLambdaAtQPs_) {
      for(i = 0; i < numPoints_; i++) {
	DELETE grdLambdaAtQPs_[i];
      }
      FREE_MEMORY(grdLambdaAtQPs_,
		  DimVec<WorldVector<double> >*,
		  numPoints_);
    }

    if(grdLambdaConst_) {
      FREE_MEMORY(grdLambdaConst_,
		  DimVec<WorldVector<double> >*,
		  numPoints_);
    }

    ::std::map<const DOFVector<double>*, VecQPInfo*>::iterator 
	it, itEnd = vecQPInfos_.end();

    for(it = vecQPInfos_.begin(); it != itEnd; ++it) {
      if(it->second->valAtQPs_) 
	FREE_MEMORY(it->second->valAtQPs_, double, numPoints_);
      if(it->second->grdAtQPs_) 
	DELETE [] it->second->grdAtQPs_;
      if(it->second->D2AtQPs_)
	DELETE [] it->second->D2AtQPs_;
      DELETE it->second;
    }
  }

  void QPInfo::initElement(const ElInfo *elInfo)
  {
    currentElInfo_ = elInfo;

    coordsNumPointsValid_ = 0;

    elementNormalNumPointsValid_ = 0;

    grdLambdaNumPointsValid_ = 0;

    ::std::map<const DOFVector<double>*, VecQPInfo*>::iterator 
	it, itEnd = vecQPInfos_.end();

    for(it = vecQPInfos_.begin(); it != itEnd; ++it) {
      it->second->valNumPointsValid_ = 0;
      it->second->grdNumPointsValid_ = 0;
      it->second->D2NumPointsValid_ = 0;
    }  
  }

  WorldVector<double> *QPInfo::getCoordsAtQPs(int numPoints) 
  {
    if(coordsNumPointsValid_ < numPoints) {
      if(!coordsAtQPs_) {
	coordsAtQPs_ = NEW WorldVector<double>[numPoints_];
      }
      int i;
      for(i = 0; i < numPoints; i++) {
	const DimVec<double>& lambda = quadrature_->getLambda(i);
	TEST_EXIT(currentElInfo_)("currentElInfo_ not set\n");
	currentElInfo_->coordToWorld(lambda, &coordsAtQPs_[i]);
      }
      coordsNumPointsValid_ = numPoints;
    }
    return coordsAtQPs_;
  }

  double *QPInfo::getVecAtQPs(const DOFVector<double>* vec, 
			      int numPoints,
			      const FastQuadrature *quadFast)
  {
    // check fast quadrature
    if(quadFast) {
      TEST_EXIT(quadrature_ == quadFast->getQuadrature())
	("quadrature_ != quadFast->quadrature\n");
    }
  
    // create new info if necessary
    if(vecQPInfos_[vec] == NULL) {
      vecQPInfos_[vec] = NEW VecQPInfo;    
    }

    VecQPInfo *localVecQPInfo = vecQPInfos_[vec];

    // update values if necessary
    if(localVecQPInfo->valNumPointsValid_ < numPoints) {

      // allocate memory if necessary
      if(localVecQPInfo->valAtQPs_ == NULL) {
	localVecQPInfo->valAtQPs_ = GET_MEMORY(double, numPoints_);
      }

      // fill memory
      vec->getVecAtQPs(currentElInfo_,
		       quadrature_,
		       quadFast,
		       localVecQPInfo->valAtQPs_,
		       numPoints);
    
      localVecQPInfo->valNumPointsValid_ = numPoints;
    }

    // return values
    return localVecQPInfo->valAtQPs_;
  }

  WorldVector<double> *QPInfo::getGrdAtQPs(const DOFVector<double>* vec, 
					   int numPoints,
					   const FastQuadrature *quadFast)
  {
    // check fast quadrature
    if(quadFast) {
      TEST_EXIT(quadrature_ == quadFast->getQuadrature())
	("quadrature_ != quadFast->quadrature\n");
    }
  
    // create new info if necessary
    if(vecQPInfos_[vec] == NULL) {
      vecQPInfos_[vec] = NEW VecQPInfo;    
    }

    VecQPInfo *localVecQPInfo = vecQPInfos_[vec];

    // update values if necessary
    if(localVecQPInfo->grdNumPointsValid_ < numPoints) {

      // allocate memory if necessary
      if(localVecQPInfo->grdAtQPs_ == NULL) {
	localVecQPInfo->grdAtQPs_ = NEW WorldVector<double>[numPoints_];
      }

      // fill memory
      vec->getGrdAtQPs(currentElInfo_,
		       quadrature_,
		       quadFast,
		       localVecQPInfo->grdAtQPs_,
		       numPoints);

      localVecQPInfo->grdNumPointsValid_ = numPoints;
    }

    // return values
    return localVecQPInfo->grdAtQPs_;
  }

  WorldMatrix<double> *QPInfo::getD2AtQPs(const DOFVector<double>* vec, 
					  int numPoints,
					  const FastQuadrature *quadFast)
  {
    // check fast quadrature
    if(quadFast) {
      TEST_EXIT(quadrature_ == quadFast->getQuadrature())
	("quadrature_ != quadFast->quadrature\n");
    }
  
    // create new info if necessary
    if(vecQPInfos_[vec] == NULL) {
      vecQPInfos_[vec] = NEW VecQPInfo;    
    }

    VecQPInfo *localVecQPInfo = vecQPInfos_[vec];

    // update values if necessary
    if(localVecQPInfo->D2NumPointsValid_ < numPoints) {

      // allocate memory if necessary
      if(localVecQPInfo->D2AtQPs_ == NULL) {
	localVecQPInfo->D2AtQPs_ = NEW WorldMatrix<double>[numPoints_];
      }

      // fill memory
      vec->getD2AtQPs(currentElInfo_,
		      quadrature_,
		      quadFast,
		      localVecQPInfo->D2AtQPs_,
		      numPoints);

      localVecQPInfo->D2NumPointsValid_ = numPoints;
    }

    // return values
    return localVecQPInfo->D2AtQPs_;
  }

  WorldVector<double> **QPInfo::getElementNormalAtQPs(int numPoints) 
  {
    TEST_EXIT(currentElInfo_)("currentElInfo_ not set\n");

    int i;
    if(currentElInfo_->getParametricOrder() > 1) {
      if(!elementNormalAtQPs_) {
	elementNormalAtQPs_ = GET_MEMORY(WorldVector<double>*,
					 numPoints_);
	for(i = 0; i < numPoints_; i++) {
	  elementNormalAtQPs_[i] = NEW WorldVector<double>;
	}
      }

      if(elementNormalNumPointsValid_ < numPoints) {
	for(i = 0; i < numPoints; i++) {
	  const DimVec<double>& lambda = quadrature_->getLambda(i);
	  currentElInfo_->getElementNormal(*(elementNormalAtQPs_[i]), &lambda);
	}
	elementNormalNumPointsValid_ = numPoints;
      }
      return elementNormalAtQPs_;
    } else {
      if(!elementNormalConst_) {
	elementNormalConst_ = GET_MEMORY(WorldVector<double>*,
					 numPoints_);
	elementNormalConst_[0] = NEW WorldVector<double>;

	for(i = 1; i < numPoints_; i++) {
	  elementNormalConst_[i] = elementNormalConst_[0];
	}
      }
      currentElInfo_->getElementNormal(*(elementNormalConst_[0]));
      return elementNormalConst_;
    }
  }

  DimVec<WorldVector<double> > **QPInfo::getGrdLambdaAtQPs(int numPoints) 
  {
    TEST_EXIT(currentElInfo_)("currentElInfo_ not set\n");

    int i;
    if(currentElInfo_->getParametricOrder() > 1) {
      if(!grdLambdaAtQPs_) {
	grdLambdaAtQPs_ = GET_MEMORY(DimVec<WorldVector<double> >*,
				     numPoints_);
	for(i = 0; i < numPoints_; i++) {
	  grdLambdaAtQPs_[i] = 
	    NEW DimVec<WorldVector<double> >(quadrature_->getDim());
	}
      }

      if(grdLambdaNumPointsValid_ < numPoints) {
	for(i = 0; i < numPoints; i++) {
	  const DimVec<double>& lambda = quadrature_->getLambda(i);
	  currentElInfo_->calcGrdLambda(*(grdLambdaAtQPs_[i]), &lambda);
	}
	grdLambdaNumPointsValid_ = numPoints;
      }
      return grdLambdaAtQPs_;
    } else {
      if(!grdLambdaConst_) {
	grdLambdaConst_ = GET_MEMORY(DimVec<WorldVector<double> >*,
				     numPoints_);
	grdLambdaConst_[0] = 
	  NEW DimVec<WorldVector<double> >(quadrature_->getDim());

	for(i = 1; i < numPoints_; i++) {
	  grdLambdaConst_[i] = grdLambdaConst_[0];
	}
      }
      currentElInfo_->calcGrdLambda(*(grdLambdaConst_[0]));
      return grdLambdaConst_;
    }
  }

  QPInfo *QPInfo::provideQPInfo(const Quadrature *quad, 
				const FastQuadrature *quadFast)
  {
    if(quadFast) {
      if(quad && (quad != quadFast->getQuadrature())) {
	ERROR_EXIT("quad != quadFast->quadrature\n");
      } else {
	quad = quadFast->getQuadrature();
      }    
    }
    if(quad) {
      if(qpInfos_[quad]) return qpInfos_[quad];
      QPInfo *newQPInfo = NEW QPInfo(quad);
      qpInfos_[quad] = newQPInfo;
      return newQPInfo;
    } else {
      return NULL;
    }
  }

  void QPInfo::clearQPInfo(const Quadrature *quad,
			   const FastQuadrature *quadFast)
  {
    if(quadFast) {
      if(quad && (quad != quadFast->getQuadrature())) {
	ERROR_EXIT("quad != quadFast->quadrature\n");
      } else {
	quad = quadFast->getQuadrature();
      }    
    }
    TEST_EXIT(quad)("no quadrature\n");
    if(qpInfos_[quad]) {
      DELETE qpInfos_[quad];
      qpInfos_.erase(quad);
    }
  }

  void QPInfo::clearAllQPInfos()
  {
    ::std::map<const Quadrature*, QPInfo*>::iterator 
      it, itEnd = qpInfos_.end();

    for(it = qpInfos_.begin(); it != itEnd; ++it) {
      DELETE it->second;
      qpInfos_.erase(it);
    }
  }

}
